"""ProductRegistryFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import hexbytes
import web3
from web3 import types
from web3.contract import contract

from .types import (
    ProductState,
    ExpirySpecification,
    ProductMetadata,
    OracleSpecification,
    BaseProduct,
    MarginSpecification,
    FuturesProductV1,
    PredictionProductV1,
)


class ProductRegistryFacet:
    """ProductRegistryFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed ProductRegistryFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    def claim_builder_rewards(
        self,
        product_id: hexbytes.HexBytes,
    ) -> contract.ContractFunction:
        """Binding for `claimBuilderRewards` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.claimBuilderRewards(
            product_id,
        )

    def expiry_specification(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> ExpirySpecification:
        """Binding for `expirySpecification` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        ExpirySpecification
        """
        return_value = self._contract.functions.expirySpecification(
            product_id,
        ).call(block_identifier=block_identifier)
        return ExpirySpecification(int(return_value[0]), int(return_value[1]))

    def futures_product_v1(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> FuturesProductV1:
        """Binding for `futuresProductV1` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        FuturesProductV1
        """
        return_value = self._contract.functions.futuresProductV1(
            product_id,
        ).call(block_identifier=block_identifier)
        return FuturesProductV1(
            BaseProduct(
                ProductMetadata(
                    eth_typing.ChecksumAddress(return_value[0][0][0]),
                    str(return_value[0][0][1]),
                    str(return_value[0][0][2]),
                ),
                OracleSpecification(
                    eth_typing.ChecksumAddress(return_value[0][1][0]),
                    int(return_value[0][1][1]),
                    int(return_value[0][1][2]),
                    int(return_value[0][1][3]),
                    hexbytes.HexBytes(return_value[0][1][4]),
                ),
                eth_typing.ChecksumAddress(return_value[0][2]),
                int(return_value[0][3]),
                int(return_value[0][4]),
                int(return_value[0][5]),
                str(return_value[0][6]),
            ),
            ExpirySpecification(int(return_value[1][0]), int(return_value[1][1])),
            MarginSpecification(int(return_value[2][0]), int(return_value[2][1])),
        )

    def id(
        self,
        product: BaseProduct,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> hexbytes.HexBytes:
        """Binding for `id` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product : BaseProduct
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.id(
            (
                (
                    product.metadata.builder,
                    product.metadata.symbol,
                    product.metadata.description,
                ),
                (
                    product.oracle_spec.oracle_address,
                    product.oracle_spec.fsv_decimals,
                    product.oracle_spec.fsp_alpha,
                    product.oracle_spec.fsp_beta,
                    product.oracle_spec.fsv_calldata,
                ),
                product.collateral_asset,
                product.start_time,
                product.point_value,
                product.price_decimals,
                product.extended_metadata,
            ),
        ).call(block_identifier=block_identifier)
        return hexbytes.HexBytes(return_value)

    def increase_builder_stake(
        self,
        product_id: hexbytes.HexBytes,
        stake_amount: int,
    ) -> contract.ContractFunction:
        """Binding for `increaseBuilderStake` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        stake_amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.increaseBuilderStake(
            product_id,
            stake_amount,
        )

    def is_builder_stake_valid(
        self,
        stake_amount: int,
        collateral: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> bool:
        """Binding for `isBuilderStakeValid` on the ProductRegistryFacet contract.

        Parameters
        ----------
        stake_amount : int
        collateral : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.isBuilderStakeValid(
            stake_amount,
            collateral,
        ).call(block_identifier=block_identifier)
        return bool(return_value)

    def min_builder_stake(
        self,
        collateral: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `minBuilderStake` on the ProductRegistryFacet contract.

        Parameters
        ----------
        collateral : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.minBuilderStake(
            collateral,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def mmr(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `mmr` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.mmr(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def prediction_product_v1(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> PredictionProductV1:
        """Binding for `predictionProductV1` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        PredictionProductV1
        """
        return_value = self._contract.functions.predictionProductV1(
            product_id,
        ).call(block_identifier=block_identifier)
        return PredictionProductV1(
            BaseProduct(
                ProductMetadata(
                    eth_typing.ChecksumAddress(return_value[0][0][0]),
                    str(return_value[0][0][1]),
                    str(return_value[0][0][2]),
                ),
                OracleSpecification(
                    eth_typing.ChecksumAddress(return_value[0][1][0]),
                    int(return_value[0][1][1]),
                    int(return_value[0][1][2]),
                    int(return_value[0][1][3]),
                    hexbytes.HexBytes(return_value[0][1][4]),
                ),
                eth_typing.ChecksumAddress(return_value[0][2]),
                int(return_value[0][3]),
                int(return_value[0][4]),
                int(return_value[0][5]),
                str(return_value[0][6]),
            ),
            ExpirySpecification(int(return_value[1][0]), int(return_value[1][1])),
            int(return_value[2]),
            int(return_value[3]),
        )

    def product_treasury(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `productTreasury` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.productTreasury(
            product_id,
        ).call(block_identifier=block_identifier)
        return eth_typing.ChecksumAddress(return_value)

    def products(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.Tuple[int, BaseProduct]:
        """Binding for `products` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        BaseProduct
        """
        return_value = self._contract.functions.products(
            product_id,
        ).call(block_identifier=block_identifier)
        return (
            int(return_value[0]),
            BaseProduct(
                ProductMetadata(
                    eth_typing.ChecksumAddress(return_value[1][0][0]),
                    str(return_value[1][0][1]),
                    str(return_value[1][0][2]),
                ),
                OracleSpecification(
                    eth_typing.ChecksumAddress(return_value[1][1][0]),
                    int(return_value[1][1][1]),
                    int(return_value[1][1][2]),
                    int(return_value[1][1][3]),
                    hexbytes.HexBytes(return_value[1][1][4]),
                ),
                eth_typing.ChecksumAddress(return_value[1][2]),
                int(return_value[1][3]),
                int(return_value[1][4]),
                int(return_value[1][5]),
                str(return_value[1][6]),
            ),
        )

    def register_future_product(
        self,
        product: FuturesProductV1,
        initial_builder_stake: int,
    ) -> contract.ContractFunction:
        """Binding for `registerFutureProduct` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product : FuturesProductV1
        initial_builder_stake : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.registerFutureProduct(
            (
                (
                    (
                        product.base.metadata.builder,
                        product.base.metadata.symbol,
                        product.base.metadata.description,
                    ),
                    (
                        product.base.oracle_spec.oracle_address,
                        product.base.oracle_spec.fsv_decimals,
                        product.base.oracle_spec.fsp_alpha,
                        product.base.oracle_spec.fsp_beta,
                        product.base.oracle_spec.fsv_calldata,
                    ),
                    product.base.collateral_asset,
                    product.base.start_time,
                    product.base.point_value,
                    product.base.price_decimals,
                    product.base.extended_metadata,
                ),
                (
                    product.expiry_spec.earliest_fsp_submission_time,
                    product.expiry_spec.tradeout_interval,
                ),
                (product.margin_spec.imr, product.margin_spec.mmr),
            ),
            initial_builder_stake,
        )

    def register_prediction_product(
        self,
        product: PredictionProductV1,
        initial_builder_stake: int,
    ) -> contract.ContractFunction:
        """Binding for `registerPredictionProduct` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product : PredictionProductV1
        initial_builder_stake : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.registerPredictionProduct(
            (
                (
                    (
                        product.base.metadata.builder,
                        product.base.metadata.symbol,
                        product.base.metadata.description,
                    ),
                    (
                        product.base.oracle_spec.oracle_address,
                        product.base.oracle_spec.fsv_decimals,
                        product.base.oracle_spec.fsp_alpha,
                        product.base.oracle_spec.fsp_beta,
                        product.base.oracle_spec.fsv_calldata,
                    ),
                    product.base.collateral_asset,
                    product.base.start_time,
                    product.base.point_value,
                    product.base.price_decimals,
                    product.base.extended_metadata,
                ),
                (
                    product.expiry_spec.earliest_fsp_submission_time,
                    product.expiry_spec.tradeout_interval,
                ),
                product.max_price,
                product.min_price,
            ),
            initial_builder_stake,
        )

    def state(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> ProductState:
        """Binding for `state` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        ProductState
        """
        return_value = self._contract.functions.state(
            product_id,
        ).call(block_identifier=block_identifier)
        return ProductState(return_value)

    def type_of(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `typeOf` on the ProductRegistryFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.typeOf(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "type": "function",
            "name": "claimBuilderRewards",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "expirySpecification",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct ExpirySpecification",
                    "components": [
                        {
                            "name": "earliestFSPSubmissionTime",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "tradeoutInterval",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "futuresProductV1",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct FuturesProductV1",
                    "components": [
                        {
                            "name": "base",
                            "type": "tuple",
                            "internalType": "struct BaseProduct",
                            "components": [
                                {
                                    "name": "metadata",
                                    "type": "tuple",
                                    "internalType": "struct ProductMetadata",
                                    "components": [
                                        {
                                            "name": "builder",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "symbol",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                        {
                                            "name": "description",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                    ],
                                },
                                {
                                    "name": "oracleSpec",
                                    "type": "tuple",
                                    "internalType": "struct OracleSpecification",
                                    "components": [
                                        {
                                            "name": "oracleAddress",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "fsvDecimals",
                                            "type": "uint8",
                                            "internalType": "uint8",
                                        },
                                        {
                                            "name": "fspAlpha",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fspBeta",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fsvCalldata",
                                            "type": "bytes",
                                            "internalType": "bytes",
                                        },
                                    ],
                                },
                                {
                                    "name": "collateralAsset",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "startTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pointValue",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "priceDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "extendedMetadata",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "expirySpec",
                            "type": "tuple",
                            "internalType": "struct ExpirySpecification",
                            "components": [
                                {
                                    "name": "earliestFSPSubmissionTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "tradeoutInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "marginSpec",
                            "type": "tuple",
                            "internalType": "struct MarginSpecification",
                            "components": [
                                {
                                    "name": "imr",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "mmr",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "id",
            "inputs": [
                {
                    "name": "product",
                    "type": "tuple",
                    "internalType": "struct BaseProduct",
                    "components": [
                        {
                            "name": "metadata",
                            "type": "tuple",
                            "internalType": "struct ProductMetadata",
                            "components": [
                                {
                                    "name": "builder",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "symbol",
                                    "type": "string",
                                    "internalType": "string",
                                },
                                {
                                    "name": "description",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "oracleSpec",
                            "type": "tuple",
                            "internalType": "struct OracleSpecification",
                            "components": [
                                {
                                    "name": "oracleAddress",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "fsvDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "fspAlpha",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "fspBeta",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "fsvCalldata",
                                    "type": "bytes",
                                    "internalType": "bytes",
                                },
                            ],
                        },
                        {
                            "name": "collateralAsset",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "startTime",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "pointValue",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "priceDecimals",
                            "type": "uint8",
                            "internalType": "uint8",
                        },
                        {
                            "name": "extendedMetadata",
                            "type": "string",
                            "internalType": "string",
                        },
                    ],
                }
            ],
            "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}],
            "stateMutability": "pure",
        },
        {
            "type": "function",
            "name": "increaseBuilderStake",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"},
                {"name": "stakeAmount", "type": "uint256", "internalType": "uint256"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "isBuilderStakeValid",
            "inputs": [
                {"name": "stakeAmount", "type": "uint256", "internalType": "uint256"},
                {"name": "collateral", "type": "address", "internalType": "address"},
            ],
            "outputs": [{"name": "", "type": "bool", "internalType": "bool"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "minBuilderStake",
            "inputs": [
                {"name": "collateral", "type": "address", "internalType": "address"}
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "mmr",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "predictionProductV1",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct PredictionProductV1",
                    "components": [
                        {
                            "name": "base",
                            "type": "tuple",
                            "internalType": "struct BaseProduct",
                            "components": [
                                {
                                    "name": "metadata",
                                    "type": "tuple",
                                    "internalType": "struct ProductMetadata",
                                    "components": [
                                        {
                                            "name": "builder",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "symbol",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                        {
                                            "name": "description",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                    ],
                                },
                                {
                                    "name": "oracleSpec",
                                    "type": "tuple",
                                    "internalType": "struct OracleSpecification",
                                    "components": [
                                        {
                                            "name": "oracleAddress",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "fsvDecimals",
                                            "type": "uint8",
                                            "internalType": "uint8",
                                        },
                                        {
                                            "name": "fspAlpha",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fspBeta",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fsvCalldata",
                                            "type": "bytes",
                                            "internalType": "bytes",
                                        },
                                    ],
                                },
                                {
                                    "name": "collateralAsset",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "startTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pointValue",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "priceDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "extendedMetadata",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "expirySpec",
                            "type": "tuple",
                            "internalType": "struct ExpirySpecification",
                            "components": [
                                {
                                    "name": "earliestFSPSubmissionTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "tradeoutInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "maxPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "minPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "productTreasury",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "products",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {"name": "", "type": "uint8", "internalType": "enum ProductType"},
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct BaseProduct",
                    "components": [
                        {
                            "name": "metadata",
                            "type": "tuple",
                            "internalType": "struct ProductMetadata",
                            "components": [
                                {
                                    "name": "builder",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "symbol",
                                    "type": "string",
                                    "internalType": "string",
                                },
                                {
                                    "name": "description",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "oracleSpec",
                            "type": "tuple",
                            "internalType": "struct OracleSpecification",
                            "components": [
                                {
                                    "name": "oracleAddress",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "fsvDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "fspAlpha",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "fspBeta",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "fsvCalldata",
                                    "type": "bytes",
                                    "internalType": "bytes",
                                },
                            ],
                        },
                        {
                            "name": "collateralAsset",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "startTime",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "pointValue",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "priceDecimals",
                            "type": "uint8",
                            "internalType": "uint8",
                        },
                        {
                            "name": "extendedMetadata",
                            "type": "string",
                            "internalType": "string",
                        },
                    ],
                },
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "registerFutureProduct",
            "inputs": [
                {
                    "name": "product",
                    "type": "tuple",
                    "internalType": "struct FuturesProductV1",
                    "components": [
                        {
                            "name": "base",
                            "type": "tuple",
                            "internalType": "struct BaseProduct",
                            "components": [
                                {
                                    "name": "metadata",
                                    "type": "tuple",
                                    "internalType": "struct ProductMetadata",
                                    "components": [
                                        {
                                            "name": "builder",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "symbol",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                        {
                                            "name": "description",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                    ],
                                },
                                {
                                    "name": "oracleSpec",
                                    "type": "tuple",
                                    "internalType": "struct OracleSpecification",
                                    "components": [
                                        {
                                            "name": "oracleAddress",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "fsvDecimals",
                                            "type": "uint8",
                                            "internalType": "uint8",
                                        },
                                        {
                                            "name": "fspAlpha",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fspBeta",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fsvCalldata",
                                            "type": "bytes",
                                            "internalType": "bytes",
                                        },
                                    ],
                                },
                                {
                                    "name": "collateralAsset",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "startTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pointValue",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "priceDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "extendedMetadata",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "expirySpec",
                            "type": "tuple",
                            "internalType": "struct ExpirySpecification",
                            "components": [
                                {
                                    "name": "earliestFSPSubmissionTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "tradeoutInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "marginSpec",
                            "type": "tuple",
                            "internalType": "struct MarginSpecification",
                            "components": [
                                {
                                    "name": "imr",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "mmr",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                    ],
                },
                {
                    "name": "initialBuilderStake",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "registerPredictionProduct",
            "inputs": [
                {
                    "name": "product",
                    "type": "tuple",
                    "internalType": "struct PredictionProductV1",
                    "components": [
                        {
                            "name": "base",
                            "type": "tuple",
                            "internalType": "struct BaseProduct",
                            "components": [
                                {
                                    "name": "metadata",
                                    "type": "tuple",
                                    "internalType": "struct ProductMetadata",
                                    "components": [
                                        {
                                            "name": "builder",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "symbol",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                        {
                                            "name": "description",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                    ],
                                },
                                {
                                    "name": "oracleSpec",
                                    "type": "tuple",
                                    "internalType": "struct OracleSpecification",
                                    "components": [
                                        {
                                            "name": "oracleAddress",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "fsvDecimals",
                                            "type": "uint8",
                                            "internalType": "uint8",
                                        },
                                        {
                                            "name": "fspAlpha",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fspBeta",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fsvCalldata",
                                            "type": "bytes",
                                            "internalType": "bytes",
                                        },
                                    ],
                                },
                                {
                                    "name": "collateralAsset",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "startTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pointValue",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "priceDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "extendedMetadata",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "expirySpec",
                            "type": "tuple",
                            "internalType": "struct ExpirySpecification",
                            "components": [
                                {
                                    "name": "earliestFSPSubmissionTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "tradeoutInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "maxPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "minPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                    ],
                },
                {
                    "name": "initialBuilderStake",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "state",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {"name": "", "type": "uint8", "internalType": "enum ProductState"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "typeOf",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {"name": "", "type": "uint8", "internalType": "enum ProductType"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "error",
            "name": "DepositNotPossible",
            "inputs": [
                {"name": "state", "type": "uint8", "internalType": "enum ProductState"}
            ],
        },
        {
            "type": "error",
            "name": "InsufficientBalance",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"},
                {"name": "balance", "type": "uint256", "internalType": "uint256"},
                {"name": "required", "type": "uint256", "internalType": "uint256"},
            ],
        },
        {
            "type": "error",
            "name": "InvalidFSPSubmissionTime",
            "inputs": [
                {"name": "startTime", "type": "uint256", "internalType": "uint256"},
                {
                    "name": "earliestFSPSubmissionTime",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
        },
        {
            "type": "error",
            "name": "InvalidFieldAccess",
            "inputs": [
                {
                    "name": "productType",
                    "type": "uint8",
                    "internalType": "enum ProductType",
                },
                {"name": "field", "type": "string", "internalType": "string"},
            ],
        },
        {
            "type": "error",
            "name": "InvalidParameter",
            "inputs": [
                {"name": "paramName", "type": "string", "internalType": "string"}
            ],
        },
        {
            "type": "error",
            "name": "InvalidPriceRange",
            "inputs": [
                {"name": "minPrice", "type": "int256", "internalType": "int256"},
                {"name": "maxPrice", "type": "int256", "internalType": "int256"},
            ],
        },
        {
            "type": "error",
            "name": "InvalidProductId",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
        },
        {
            "type": "error",
            "name": "InvalidStartTime",
            "inputs": [
                {"name": "startTime", "type": "uint256", "internalType": "uint256"},
                {
                    "name": "blockTimestamp",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
        },
        {
            "type": "error",
            "name": "NotEnoughFee",
            "inputs": [
                {"name": "fee", "type": "uint256", "internalType": "uint256"},
                {"name": "requiredFee", "type": "uint256", "internalType": "uint256"},
            ],
        },
        {
            "type": "error",
            "name": "NotFound",
            "inputs": [
                {"name": "parameter", "type": "string", "internalType": "string"}
            ],
        },
        {
            "type": "error",
            "name": "NotImplemented",
            "inputs": [{"name": "feature", "type": "string", "internalType": "string"}],
        },
        {
            "type": "error",
            "name": "ProductExists",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
        },
        {
            "type": "error",
            "name": "SafeCastOverflowedUintToInt",
            "inputs": [{"name": "value", "type": "uint256", "internalType": "uint256"}],
        },
        {
            "type": "error",
            "name": "Unauthorized",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"}
            ],
        },
    ],
)



"""ERC20 contract binding and data structures."""

import typing

import eth_typing
import web3
from web3.contract import contract


class ERC20:
    """ERC20 contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed ERC20 contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def Approval(self) -> contract.ContractEvent:
        """Binding for `event Approval` on the ERC20 contract."""
        return self._contract.events.Approval

    @property
    def Transfer(self) -> contract.ContractEvent:
        """Binding for `event Transfer` on the ERC20 contract."""
        return self._contract.events.Transfer

    def allowance(
        self,
        owner: eth_typing.ChecksumAddress,
        spender: eth_typing.ChecksumAddress,
    ) -> int:
        """Binding for `allowance` on the ERC20 contract.

        Parameters
        ----------
        owner : eth_typing.ChecksumAddress
        spender : eth_typing.ChecksumAddress

        Returns
        -------
        int
        """
        return_value = self._contract.functions.allowance(
            owner,
            spender,
        ).call()
        return int(return_value)

    def approve(
        self,
        spender: eth_typing.ChecksumAddress,
        value: int,
    ) -> contract.ContractFunction:
        """Binding for `approve` on the ERC20 contract.

        Parameters
        ----------
        spender : eth_typing.ChecksumAddress
        value : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.approve(
            spender,
            value,
        )

    def balance_of(
        self,
        account: eth_typing.ChecksumAddress,
    ) -> int:
        """Binding for `balanceOf` on the ERC20 contract.

        Parameters
        ----------
        account : eth_typing.ChecksumAddress

        Returns
        -------
        int
        """
        return_value = self._contract.functions.balanceOf(
            account,
        ).call()
        return int(return_value)

    def decimals(
        self,
    ) -> int:
        """Binding for `decimals` on the ERC20 contract.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.decimals().call()
        return int(return_value)

    def name(
        self,
    ) -> str:
        """Binding for `name` on the ERC20 contract.

        Returns
        -------
        str
        """
        return_value = self._contract.functions.name().call()
        return str(return_value)

    def symbol(
        self,
    ) -> str:
        """Binding for `symbol` on the ERC20 contract.

        Returns
        -------
        str
        """
        return_value = self._contract.functions.symbol().call()
        return str(return_value)

    def total_supply(
        self,
    ) -> int:
        """Binding for `totalSupply` on the ERC20 contract.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.totalSupply().call()
        return int(return_value)

    def transfer(
        self,
        to: eth_typing.ChecksumAddress,
        value: int,
    ) -> contract.ContractFunction:
        """Binding for `transfer` on the ERC20 contract.

        Parameters
        ----------
        to : eth_typing.ChecksumAddress
        value : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.transfer(
            to,
            value,
        )

    def transfer_from(
        self,
        from_: eth_typing.ChecksumAddress,
        to: eth_typing.ChecksumAddress,
        value: int,
    ) -> contract.ContractFunction:
        """Binding for `transferFrom` on the ERC20 contract.

        Parameters
        ----------
        from_ : eth_typing.ChecksumAddress
        to : eth_typing.ChecksumAddress
        value : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.transferFrom(
            from_,
            to,
            value,
        )


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "inputs": [
                {"internalType": "address", "name": "spender", "type": "address"},
                {"internalType": "uint256", "name": "allowance", "type": "uint256"},
                {"internalType": "uint256", "name": "needed", "type": "uint256"},
            ],
            "name": "ERC20InsufficientAllowance",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "sender", "type": "address"},
                {"internalType": "uint256", "name": "balance", "type": "uint256"},
                {"internalType": "uint256", "name": "needed", "type": "uint256"},
            ],
            "name": "ERC20InsufficientBalance",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "approver", "type": "address"}
            ],
            "name": "ERC20InvalidApprover",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "receiver", "type": "address"}
            ],
            "name": "ERC20InvalidReceiver",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "sender", "type": "address"}
            ],
            "name": "ERC20InvalidSender",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "spender", "type": "address"}
            ],
            "name": "ERC20InvalidSpender",
            "type": "error",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "owner",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "spender",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "value",
                    "type": "uint256",
                },
            ],
            "name": "Approval",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "from",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "to",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "value",
                    "type": "uint256",
                },
            ],
            "name": "Transfer",
            "type": "event",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "owner", "type": "address"},
                {"internalType": "address", "name": "spender", "type": "address"},
            ],
            "name": "allowance",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "spender", "type": "address"},
                {"internalType": "uint256", "name": "value", "type": "uint256"},
            ],
            "name": "approve",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "account", "type": "address"}
            ],
            "name": "balanceOf",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "decimals",
            "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "name",
            "outputs": [{"internalType": "string", "name": "", "type": "string"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "symbol",
            "outputs": [{"internalType": "string", "name": "", "type": "string"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "totalSupply",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "to", "type": "address"},
                {"internalType": "uint256", "name": "value", "type": "uint256"},
            ],
            "name": "transfer",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "from", "type": "address"},
                {"internalType": "address", "name": "to", "type": "address"},
                {"internalType": "uint256", "name": "value", "type": "uint256"},
            ],
            "name": "transferFrom",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "nonpayable",
            "type": "function",
        },
    ],
)



from os import path

import pytest

import afp
from afp.constants import defaults
from afp.exceptions import ConfigurationError


def test_AFP_constructor__creates_KeyFileAuthenticator_from_defaults(
    monkeypatch,
):
    keyfile = path.join(path.dirname(__file__), "assets", "test.key")
    monkeypatch.setattr(defaults, "KEYFILE", keyfile)
    monkeypatch.setattr(defaults, "KEYFILE_PASSWORD", "foobar")

    app = afp.AFP()
    assert isinstance(app.config.authenticator, afp.KeyfileAuthenticator)
    assert (
        app.config.authenticator.address == "0xFfbf2643CF22760AfD3b878BA8aE849c48944Aa5"
    )


def test_AFP_constructor__creates_PrivateKeyAuthenticator_from_defaults(
    monkeypatch,
):
    private_key = "0x93f2348b2b890be53f55344b7b118b53778440eea0dc84a6f86fb6e7f6bd77a6"
    monkeypatch.setattr(defaults, "PRIVATE_KEY", private_key)

    app = afp.AFP()
    assert isinstance(app.config.authenticator, afp.PrivateKeyAuthenticator)
    assert (
        app.config.authenticator.address == "0xFfbf2643CF22760AfD3b878BA8aE849c48944Aa5"
    )


def test_AFP_constructor__raises_error_for_ambiguous_defaults(monkeypatch):
    monkeypatch.setattr(defaults, "KEYFILE", "baz")
    monkeypatch.setattr(defaults, "PRIVATE_KEY", "0x")

    with pytest.raises(ConfigurationError):
        afp.AFP()



"""Example for creating a product from JSON specification."""

import os
from decimal import Decimal
from os import path
from pprint import pprint

import afp


AUTONITY_RPC_URL = "https://bakerloo.autonity-apis.com"
IPFS_API_URL = "https://rpc.filebase.io"

PRIVATE_KEY = os.environ["PRIVATE_KEY"]
IPFS_API_KEY = os.environ["IPFS_API_KEY"]


def main():
    # 0. Configure the SDK
    authenticator = afp.PrivateKeyAuthenticator(PRIVATE_KEY)
    app = afp.AFP(
        authenticator=authenticator,
        rpc_url=AUTONITY_RPC_URL,
        ipfs_api_url=IPFS_API_URL,
        ipfs_api_key=IPFS_API_KEY,
    )
    product = app.Product()

    # 1. Validate the JSON specification
    with open(
        path.join(path.dirname(__file__), "prediction-product-example.json")
    ) as f:
        specification = product.validate_json(f.read())

    # 2. Upload the extended metadata to IPFS
    pinned_specification = product.pin(specification)
    pprint(product.dump(pinned_specification))

    # 3. Send registration transaction
    tx = product.register(pinned_specification, initial_builder_stake=Decimal("0"))
    print(tx.receipt)

    # 4. Check product state
    id = product.id(pinned_specification)
    print(product.state(id))

    # 5. Download the product specification from blockchain & IPFS and re-validate it
    downloaded_specification = product.get(id)

    # 6. Write the downloaded product specification into a JSON file
    with open("downloaded-product-example.json", "w") as f:
        f.write(product.dump_json(downloaded_specification))


if __name__ == "__main__":
    main()



"""AFP data structures."""

from decimal import Decimal
from itertools import chain
from typing import Annotated, Any, ClassVar, Literal, Self

from pydantic import AfterValidator, BeforeValidator, Field, model_validator

from . import constants, validators
from .constants import schema_cids
from .enums import ListingState, OrderSide, OrderState, OrderType, TradeState
from .types import (
    CID,
    URL,
    AliasedModel,
    ISODate,
    ISODateTime,
    Model,
    PinnedModel,
    Timestamp,
)


# Trading API


class ExchangeProduct(AliasedModel):
    id: str
    symbol: Annotated[str, AfterValidator(validators.validate_all_caps)]
    tick_size: int
    collateral_asset: str
    listing_state: ListingState
    min_price: Decimal
    max_price: Decimal

    def __str__(self) -> str:
        return self.id


class IntentData(AliasedModel):
    trading_protocol_id: str
    product_id: str
    limit_price: Decimal
    quantity: Annotated[int, Field(gt=0)]
    max_trading_fee_rate: Annotated[
        Decimal,
        Field(le=Decimal((2**32 - 1) / constants.FEE_RATE_MULTIPLIER)),  # uint32
    ]
    side: OrderSide
    good_until_time: Timestamp
    nonce: int
    referral: Annotated[str, AfterValidator(validators.validate_address)]


class Intent(AliasedModel):
    hash: str
    margin_account_id: str
    intent_account_id: str
    signature: str
    data: IntentData


class Order(AliasedModel):
    id: str
    type: OrderType
    timestamp: Timestamp
    state: OrderState
    fill_quantity: int
    intent: Intent


class OrderCancellationData(AliasedModel):
    intent_hash: Annotated[str, AfterValidator(validators.validate_hexstr32)]
    nonce: int
    intent_account_id: str
    signature: str


class Trade(AliasedModel):
    # Convert ID from int to str for backward compatibility
    id: Annotated[str, BeforeValidator(str)]
    product_id: str
    price: Decimal
    timestamp: Timestamp
    state: TradeState
    transaction_id: str | None
    rejection_reason: str | None


class OrderFill(AliasedModel):
    order: Order
    trade: Trade
    quantity: int
    price: Decimal
    trading_fee_rate: Decimal


class MarketDepthItem(AliasedModel):
    price: Decimal
    quantity: int


class MarketDepthData(AliasedModel):
    product_id: str
    bids: list[MarketDepthItem]
    asks: list[MarketDepthItem]


class OHLCVItem(AliasedModel):
    timestamp: Timestamp
    open: Decimal
    high: Decimal
    low: Decimal
    close: Decimal
    volume: int


# Margin Account API


class Transaction(Model):
    hash: str
    data: dict[str, Any]
    receipt: dict[str, Any]


class Position(Model):
    id: str
    quantity: int
    cost_basis: Decimal
    maintenance_margin: Decimal
    pnl: Decimal


# Product API


class ExpirySpecification(AliasedModel):
    earliest_fsp_submission_time: Annotated[
        ISODateTime, Field(alias="earliestFSPSubmissionTime")
    ]
    tradeout_interval: Annotated[int, Field(ge=0)]


class OracleSpecification(AliasedModel):
    oracle_address: Annotated[str, AfterValidator(validators.validate_address)]
    fsv_decimals: Annotated[int, Field(ge=0, le=255)]  # uint8
    fsp_alpha: Decimal
    fsp_beta: Decimal
    fsv_calldata: Annotated[str, AfterValidator(validators.validate_hexstr)]


class ProductMetadata(AliasedModel):
    builder: Annotated[str, AfterValidator(validators.validate_address)]
    symbol: Annotated[
        str,
        Field(pattern=r"^[A-Z0-9]{1,16}$", min_length=1, max_length=16),
    ]
    description: str


class BaseProduct(AliasedModel):
    metadata: ProductMetadata
    oracle_spec: OracleSpecification
    collateral_asset: Annotated[str, AfterValidator(validators.validate_address)]
    start_time: ISODateTime
    point_value: Decimal
    price_decimals: Annotated[int, Field(ge=0, le=255)]  # uint8
    extended_metadata: CID | None = None


class PredictionProductV1(AliasedModel):
    base: BaseProduct
    expiry_spec: ExpirySpecification
    min_price: Decimal
    max_price: Decimal

    @model_validator(mode="after")
    def _cross_validate(self) -> Self:
        validators.validate_price_limits(self.min_price, self.max_price)
        validators.validate_time_limits(
            self.base.start_time, self.expiry_spec.earliest_fsp_submission_time
        )
        return self


# Extended metadata schemas


class ApiSpec(Model):
    standard: Literal["JSONPath", "GraphQL"]
    spec_variant: Literal["underlying-history", "product-fsv"] | None = None


class ApiSpecJSONPath(ApiSpec):
    standard: Literal["JSONPath"] = "JSONPath"  # type: ignore
    url: URL
    date_path: str
    value_path: str
    auth_param_location: Literal["query", "header", "none"] = "none"
    auth_param_name: str | None = None
    auth_param_prefix: str | None = None
    continuation_token_param: str | None = None
    continuation_token_path: str | None = None
    date_format_custom: str | None = None
    date_format_type: Literal["iso_8601", "unix_timestamp", "custom"] = "iso_8601"
    headers: dict[str, str] | None = None
    max_pages: Annotated[int | None, Field(ge=1)] = 10
    timestamp_scale: Annotated[int | float, Field(ge=1)] = 1
    timezone: Annotated[
        str,
        Field(pattern=r"^[A-Za-z][A-Za-z0-9_+-]*(/[A-Za-z][A-Za-z0-9_+-]*)*$"),
    ] = "UTC"


class BaseCaseResolution(Model):
    condition: Annotated[str, Field(min_length=1)]
    fsp_resolution: Annotated[str, Field(min_length=1)]


class EdgeCase(Model):
    condition: Annotated[str, Field(min_length=1)]
    fsp_resolution: Annotated[str, Field(min_length=1)]


class OutcomeSpace(PinnedModel):
    SCHEMA_CID: ClassVar[CID] = schema_cids.OUTCOME_SPACE_V020

    fsp_type: Literal["scalar", "binary", "ternary"]
    description: Annotated[str, Field(min_length=1)]
    base_case: BaseCaseResolution
    edge_cases: Annotated[list[EdgeCase], Field(default_factory=list)]


class OutcomeSpaceScalar(OutcomeSpace):
    SCHEMA_CID: ClassVar[CID] = schema_cids.OUTCOME_SPACE_SCALAR_V020

    fsp_type: Literal["scalar"] = "scalar"  # type: ignore
    units: Annotated[str, Field(min_length=1)]
    source_name: Annotated[str, Field(min_length=1)]
    source_uri: URL


class OutcomeSpaceTimeSeries(OutcomeSpaceScalar):
    SCHEMA_CID: ClassVar[CID] = schema_cids.OUTCOME_SPACE_TIME_SERIES_V020

    frequency: Literal[
        "daily",
        "weekly",
        "fortnightly",
        "semimonthly",
        "monthly",
        "quarterly",
        "yearly",
    ]
    history_api_spec: ApiSpecJSONPath | ApiSpec | None = None


class TemporalObservation(Model):
    reference_date: ISODate
    release_date: ISODate


class OutcomePoint(PinnedModel):
    SCHEMA_CID: ClassVar[CID] = schema_cids.OUTCOME_POINT_V020

    fsp_type: Literal["scalar", "binary", "ternary"]


class OutcomePointTimeSeries(OutcomePoint):
    SCHEMA_CID: ClassVar[CID] = schema_cids.OUTCOME_POINT_TIME_SERIES_V020

    fsp_type: Literal["scalar"] = "scalar"  # type: ignore
    observation: TemporalObservation


class OutcomePointEvent(OutcomePoint):
    SCHEMA_CID: ClassVar[CID] = schema_cids.OUTCOME_POINT_EVENT_V020

    fsp_type: Literal["binary", "ternary"]  # type: ignore
    outcome: Annotated[str, Field(min_length=1)]


class OracleConfig(PinnedModel):
    SCHEMA_CID: ClassVar[CID] = schema_cids.ORACLE_CONFIG_V020

    description: Annotated[str, Field(min_length=1)]
    project_url: URL | None = None


class OracleConfigPrototype1(OracleConfig):
    SCHEMA_CID: ClassVar[CID] = schema_cids.ORACLE_CONFIG_PROTOTYPE1_V020

    evaluation_api_spec: ApiSpecJSONPath | ApiSpec


class OracleFallback(PinnedModel):
    SCHEMA_CID: ClassVar[CID] = schema_cids.ORACLE_FALLBACK_V020

    fallback_time: ISODateTime
    fallback_fsp: Decimal


class PredictionProduct(Model):
    product: PredictionProductV1
    outcome_space: OutcomeSpaceTimeSeries | OutcomeSpaceScalar | OutcomeSpace
    outcome_point: OutcomePointEvent | OutcomePointTimeSeries | OutcomePoint
    oracle_config: OracleConfigPrototype1 | OracleConfig
    oracle_fallback: OracleFallback

    @model_validator(mode="after")
    def _cross_validate(self) -> Self:
        validators.validate_matching_fsp_types(
            self.outcome_space.fsp_type, self.outcome_point.fsp_type
        )
        validators.validate_oracle_fallback_time(
            self.oracle_fallback.fallback_time,
            self.product.expiry_spec.earliest_fsp_submission_time,
        )
        validators.validate_oracle_fallback_fsp(
            self.oracle_fallback.fallback_fsp,
            self.product.min_price,
            self.product.max_price,
        )
        validators.validate_outcome_space_template_variables(
            [
                self.outcome_space.base_case.condition,
                self.outcome_space.base_case.fsp_resolution,
            ]
            + list(
                chain.from_iterable(
                    [edge_case.condition, edge_case.fsp_resolution]
                    for edge_case in self.outcome_space.edge_cases
                )
            ),
            self.outcome_point.model_dump(),
        )
        if isinstance(self.outcome_space, OutcomeSpaceTimeSeries) and isinstance(
            self.outcome_point, OutcomePointTimeSeries
        ):
            validators.validate_symbol(
                self.product.base.metadata.symbol,
                self.outcome_space.frequency,
                self.outcome_point.observation.release_date,
            )
        return self



"""ProductRegistry contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import hexbytes
import web3
from web3 import types
from web3.contract import contract

from .types import (
    ProductState,
    ExpirySpecification,
    PendingStakeData,
    BuilderStakeData,
    ProductMetadata,
    OracleSpecification,
    BaseProduct,
    MarginSpecification,
    FuturesProductV1,
    PredictionProductV1,
)


class ProductRegistry:
    """ProductRegistry contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed ProductRegistry contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def Initialized(self) -> contract.ContractEvent:
        """Binding for `event Initialized` on the ProductRegistry contract."""
        return self._contract.events.Initialized

    @property
    def OwnershipTransferred(self) -> contract.ContractEvent:
        """Binding for `event OwnershipTransferred` on the ProductRegistry contract."""
        return self._contract.events.OwnershipTransferred

    @property
    def ProductRegistered(self) -> contract.ContractEvent:
        """Binding for `event ProductRegistered` on the ProductRegistry contract."""
        return self._contract.events.ProductRegistered

    @property
    def Upgraded(self) -> contract.ContractEvent:
        """Binding for `event Upgraded` on the ProductRegistry contract."""
        return self._contract.events.Upgraded

    def upgrade_interface_version(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> str:
        """Binding for `UPGRADE_INTERFACE_VERSION` on the ProductRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        str
        """
        return_value = self._contract.functions.UPGRADE_INTERFACE_VERSION().call(
            block_identifier=block_identifier
        )
        return str(return_value)

    def add_allowed(
        self,
        addr: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `addAllowed` on the ProductRegistry contract.

        Parameters
        ----------
        addr : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.addAllowed(
            addr,
        )

    def admin(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `admin` on the ProductRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.admin().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def allowed(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[eth_typing.ChecksumAddress]:
        """Binding for `allowed` on the ProductRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[eth_typing.ChecksumAddress]
        """
        return_value = self._contract.functions.allowed().call(
            block_identifier=block_identifier
        )
        return [
            eth_typing.ChecksumAddress(return_value_elem)
            for return_value_elem in return_value
        ]

    def builder_stake_data(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> BuilderStakeData:
        """Binding for `builderStakeData` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        BuilderStakeData
        """
        return_value = self._contract.functions.builderStakeData(
            product_id,
        ).call(block_identifier=block_identifier)
        return BuilderStakeData(
            int(return_value[0]),
            int(return_value[1]),
            PendingStakeData(int(return_value[2][0]), int(return_value[2][1])),
        )

    def claim_builder_rewards(
        self,
        product_id: hexbytes.HexBytes,
    ) -> contract.ContractFunction:
        """Binding for `claimBuilderRewards` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.claimBuilderRewards(
            product_id,
        )

    def claimable_builder_rewards(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.Tuple[int, int]:
        """Binding for `claimableBuilderRewards` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        int
        """
        return_value = self._contract.functions.claimableBuilderRewards(
            product_id,
        ).call(block_identifier=block_identifier)
        return (
            int(return_value[0]),
            int(return_value[1]),
        )

    def clearing(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `clearing` on the ProductRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.clearing().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def collateral_asset(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `collateralAsset` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.collateralAsset(
            product_id,
        ).call(block_identifier=block_identifier)
        return eth_typing.ChecksumAddress(return_value)

    def earliest_fsp_submission_time(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `earliestFSPSubmissionTime` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.earliestFSPSubmissionTime(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def futures_product_v1(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> FuturesProductV1:
        """Binding for `futuresProductV1` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        FuturesProductV1
        """
        return_value = self._contract.functions.futuresProductV1(
            product_id,
        ).call(block_identifier=block_identifier)
        return FuturesProductV1(
            BaseProduct(
                ProductMetadata(
                    eth_typing.ChecksumAddress(return_value[0][0][0]),
                    str(return_value[0][0][1]),
                    str(return_value[0][0][2]),
                ),
                OracleSpecification(
                    eth_typing.ChecksumAddress(return_value[0][1][0]),
                    int(return_value[0][1][1]),
                    int(return_value[0][1][2]),
                    int(return_value[0][1][3]),
                    hexbytes.HexBytes(return_value[0][1][4]),
                ),
                eth_typing.ChecksumAddress(return_value[0][2]),
                int(return_value[0][3]),
                int(return_value[0][4]),
                int(return_value[0][5]),
                str(return_value[0][6]),
            ),
            ExpirySpecification(int(return_value[1][0]), int(return_value[1][1])),
            MarginSpecification(int(return_value[2][0]), int(return_value[2][1])),
        )

    def id(
        self,
        product: BaseProduct,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> hexbytes.HexBytes:
        """Binding for `id` on the ProductRegistry contract.

        Parameters
        ----------
        product : BaseProduct
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.id(
            (
                (
                    product.metadata.builder,
                    product.metadata.symbol,
                    product.metadata.description,
                ),
                (
                    product.oracle_spec.oracle_address,
                    product.oracle_spec.fsv_decimals,
                    product.oracle_spec.fsp_alpha,
                    product.oracle_spec.fsp_beta,
                    product.oracle_spec.fsv_calldata,
                ),
                product.collateral_asset,
                product.start_time,
                product.point_value,
                product.price_decimals,
                product.extended_metadata,
            ),
        ).call(block_identifier=block_identifier)
        return hexbytes.HexBytes(return_value)

    def increase_builder_stake(
        self,
        product_id: hexbytes.HexBytes,
        fee: int,
    ) -> contract.ContractFunction:
        """Binding for `increaseBuilderStake` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        fee : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.increaseBuilderStake(
            product_id,
            fee,
        )

    def initialize(
        self,
        clearing_: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `initialize` on the ProductRegistry contract.

        Parameters
        ----------
        clearing_ : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initialize(
            clearing_,
        )

    def max_price(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `maxPrice` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.maxPrice(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def min_price(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `minPrice` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.minPrice(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def mmr(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `mmr` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.mmr(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def next_interval(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `nextInterval` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.nextInterval(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def oracle_specification(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> OracleSpecification:
        """Binding for `oracleSpecification` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        OracleSpecification
        """
        return_value = self._contract.functions.oracleSpecification(
            product_id,
        ).call(block_identifier=block_identifier)
        return OracleSpecification(
            eth_typing.ChecksumAddress(return_value[0]),
            int(return_value[1]),
            int(return_value[2]),
            int(return_value[3]),
            hexbytes.HexBytes(return_value[4]),
        )

    def owner(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `owner` on the ProductRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.owner().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def point_value(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `pointValue` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.pointValue(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def prediction_product_v1(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> PredictionProductV1:
        """Binding for `predictionProductV1` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        PredictionProductV1
        """
        return_value = self._contract.functions.predictionProductV1(
            product_id,
        ).call(block_identifier=block_identifier)
        return PredictionProductV1(
            BaseProduct(
                ProductMetadata(
                    eth_typing.ChecksumAddress(return_value[0][0][0]),
                    str(return_value[0][0][1]),
                    str(return_value[0][0][2]),
                ),
                OracleSpecification(
                    eth_typing.ChecksumAddress(return_value[0][1][0]),
                    int(return_value[0][1][1]),
                    int(return_value[0][1][2]),
                    int(return_value[0][1][3]),
                    hexbytes.HexBytes(return_value[0][1][4]),
                ),
                eth_typing.ChecksumAddress(return_value[0][2]),
                int(return_value[0][3]),
                int(return_value[0][4]),
                int(return_value[0][5]),
                str(return_value[0][6]),
            ),
            ExpirySpecification(int(return_value[1][0]), int(return_value[1][1])),
            int(return_value[2]),
            int(return_value[3]),
        )

    def price_decimals(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `priceDecimals` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.priceDecimals(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def product_treasury(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `productTreasury` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.productTreasury(
            product_id,
        ).call(block_identifier=block_identifier)
        return eth_typing.ChecksumAddress(return_value)

    def products(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.Tuple[int, BaseProduct]:
        """Binding for `products` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        BaseProduct
        """
        return_value = self._contract.functions.products(
            product_id,
        ).call(block_identifier=block_identifier)
        return (
            int(return_value[0]),
            BaseProduct(
                ProductMetadata(
                    eth_typing.ChecksumAddress(return_value[1][0][0]),
                    str(return_value[1][0][1]),
                    str(return_value[1][0][2]),
                ),
                OracleSpecification(
                    eth_typing.ChecksumAddress(return_value[1][1][0]),
                    int(return_value[1][1][1]),
                    int(return_value[1][1][2]),
                    int(return_value[1][1][3]),
                    hexbytes.HexBytes(return_value[1][1][4]),
                ),
                eth_typing.ChecksumAddress(return_value[1][2]),
                int(return_value[1][3]),
                int(return_value[1][4]),
                int(return_value[1][5]),
                str(return_value[1][6]),
            ),
        )

    def proxiable_uuid(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> hexbytes.HexBytes:
        """Binding for `proxiableUUID` on the ProductRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.proxiableUUID().call(
            block_identifier=block_identifier
        )
        return hexbytes.HexBytes(return_value)

    def register_future_product(
        self,
        product: FuturesProductV1,
        initial_builder_stake: int,
    ) -> contract.ContractFunction:
        """Binding for `registerFutureProduct` on the ProductRegistry contract.

        Parameters
        ----------
        product : FuturesProductV1
        initial_builder_stake : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.registerFutureProduct(
            (
                (
                    (
                        product.base.metadata.builder,
                        product.base.metadata.symbol,
                        product.base.metadata.description,
                    ),
                    (
                        product.base.oracle_spec.oracle_address,
                        product.base.oracle_spec.fsv_decimals,
                        product.base.oracle_spec.fsp_alpha,
                        product.base.oracle_spec.fsp_beta,
                        product.base.oracle_spec.fsv_calldata,
                    ),
                    product.base.collateral_asset,
                    product.base.start_time,
                    product.base.point_value,
                    product.base.price_decimals,
                    product.base.extended_metadata,
                ),
                (
                    product.expiry_spec.earliest_fsp_submission_time,
                    product.expiry_spec.tradeout_interval,
                ),
                (product.margin_spec.imr, product.margin_spec.mmr),
            ),
            initial_builder_stake,
        )

    def register_prediction_product(
        self,
        product: PredictionProductV1,
        initial_builder_stake: int,
    ) -> contract.ContractFunction:
        """Binding for `registerPredictionProduct` on the ProductRegistry contract.

        Parameters
        ----------
        product : PredictionProductV1
        initial_builder_stake : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.registerPredictionProduct(
            (
                (
                    (
                        product.base.metadata.builder,
                        product.base.metadata.symbol,
                        product.base.metadata.description,
                    ),
                    (
                        product.base.oracle_spec.oracle_address,
                        product.base.oracle_spec.fsv_decimals,
                        product.base.oracle_spec.fsp_alpha,
                        product.base.oracle_spec.fsp_beta,
                        product.base.oracle_spec.fsv_calldata,
                    ),
                    product.base.collateral_asset,
                    product.base.start_time,
                    product.base.point_value,
                    product.base.price_decimals,
                    product.base.extended_metadata,
                ),
                (
                    product.expiry_spec.earliest_fsp_submission_time,
                    product.expiry_spec.tradeout_interval,
                ),
                product.max_price,
                product.min_price,
            ),
            initial_builder_stake,
        )

    def register_prediction_product_for(
        self,
        product: PredictionProductV1,
        initial_builder_stake: int,
    ) -> contract.ContractFunction:
        """Binding for `registerPredictionProductFor` on the ProductRegistry contract.

        Parameters
        ----------
        product : PredictionProductV1
        initial_builder_stake : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.registerPredictionProductFor(
            (
                (
                    (
                        product.base.metadata.builder,
                        product.base.metadata.symbol,
                        product.base.metadata.description,
                    ),
                    (
                        product.base.oracle_spec.oracle_address,
                        product.base.oracle_spec.fsv_decimals,
                        product.base.oracle_spec.fsp_alpha,
                        product.base.oracle_spec.fsp_beta,
                        product.base.oracle_spec.fsv_calldata,
                    ),
                    product.base.collateral_asset,
                    product.base.start_time,
                    product.base.point_value,
                    product.base.price_decimals,
                    product.base.extended_metadata,
                ),
                (
                    product.expiry_spec.earliest_fsp_submission_time,
                    product.expiry_spec.tradeout_interval,
                ),
                product.max_price,
                product.min_price,
            ),
            initial_builder_stake,
        )

    def remove_allowed(
        self,
        addr: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `removeAllowed` on the ProductRegistry contract.

        Parameters
        ----------
        addr : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.removeAllowed(
            addr,
        )

    def renounce_ownership(
        self,
    ) -> contract.ContractFunction:
        """Binding for `renounceOwnership` on the ProductRegistry contract.

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.renounceOwnership()

    def state(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> ProductState:
        """Binding for `state` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        ProductState
        """
        return_value = self._contract.functions.state(
            product_id,
        ).call(block_identifier=block_identifier)
        return ProductState(return_value)

    def tick_size(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `tickSize` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.tickSize(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def transfer_ownership(
        self,
        new_owner: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `transferOwnership` on the ProductRegistry contract.

        Parameters
        ----------
        new_owner : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.transferOwnership(
            new_owner,
        )

    def upgrade_to_and_call(
        self,
        new_implementation: eth_typing.ChecksumAddress,
        data: hexbytes.HexBytes,
    ) -> contract.ContractFunction:
        """Binding for `upgradeToAndCall` on the ProductRegistry contract.

        Parameters
        ----------
        new_implementation : eth_typing.ChecksumAddress
        data : hexbytes.HexBytes

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.upgradeToAndCall(
            new_implementation,
            data,
        )


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "type": "function",
            "name": "UPGRADE_INTERFACE_VERSION",
            "inputs": [],
            "outputs": [{"name": "", "type": "string", "internalType": "string"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "addAllowed",
            "inputs": [{"name": "addr", "type": "address", "internalType": "address"}],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "admin",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "allowed",
            "inputs": [],
            "outputs": [{"name": "", "type": "address[]", "internalType": "address[]"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "builderStakeData",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct BuilderStakeData",
                    "components": [
                        {
                            "name": "payoutRemaining",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "claimableRewards",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "pendingStakeData",
                            "type": "tuple",
                            "internalType": "struct PendingStakeData",
                            "components": [
                                {
                                    "name": "payout",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "activeAt",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "claimBuilderRewards",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "claimableBuilderRewards",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {
                    "name": "builderRewards",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "treasuryRewads",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "clearing",
            "inputs": [],
            "outputs": [
                {
                    "name": "",
                    "type": "address",
                    "internalType": "contract IProductRegistryFacet",
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "collateralAsset",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "earliestFSPSubmissionTime",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "futuresProductV1",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct FuturesProductV1",
                    "components": [
                        {
                            "name": "base",
                            "type": "tuple",
                            "internalType": "struct BaseProduct",
                            "components": [
                                {
                                    "name": "metadata",
                                    "type": "tuple",
                                    "internalType": "struct ProductMetadata",
                                    "components": [
                                        {
                                            "name": "builder",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "symbol",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                        {
                                            "name": "description",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                    ],
                                },
                                {
                                    "name": "oracleSpec",
                                    "type": "tuple",
                                    "internalType": "struct OracleSpecification",
                                    "components": [
                                        {
                                            "name": "oracleAddress",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "fsvDecimals",
                                            "type": "uint8",
                                            "internalType": "uint8",
                                        },
                                        {
                                            "name": "fspAlpha",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fspBeta",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fsvCalldata",
                                            "type": "bytes",
                                            "internalType": "bytes",
                                        },
                                    ],
                                },
                                {
                                    "name": "collateralAsset",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "startTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pointValue",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "priceDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "extendedMetadata",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "expirySpec",
                            "type": "tuple",
                            "internalType": "struct ExpirySpecification",
                            "components": [
                                {
                                    "name": "earliestFSPSubmissionTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "tradeoutInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "marginSpec",
                            "type": "tuple",
                            "internalType": "struct MarginSpecification",
                            "components": [
                                {
                                    "name": "imr",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "mmr",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "id",
            "inputs": [
                {
                    "name": "product",
                    "type": "tuple",
                    "internalType": "struct BaseProduct",
                    "components": [
                        {
                            "name": "metadata",
                            "type": "tuple",
                            "internalType": "struct ProductMetadata",
                            "components": [
                                {
                                    "name": "builder",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "symbol",
                                    "type": "string",
                                    "internalType": "string",
                                },
                                {
                                    "name": "description",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "oracleSpec",
                            "type": "tuple",
                            "internalType": "struct OracleSpecification",
                            "components": [
                                {
                                    "name": "oracleAddress",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "fsvDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "fspAlpha",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "fspBeta",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "fsvCalldata",
                                    "type": "bytes",
                                    "internalType": "bytes",
                                },
                            ],
                        },
                        {
                            "name": "collateralAsset",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "startTime",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "pointValue",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "priceDecimals",
                            "type": "uint8",
                            "internalType": "uint8",
                        },
                        {
                            "name": "extendedMetadata",
                            "type": "string",
                            "internalType": "string",
                        },
                    ],
                }
            ],
            "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "increaseBuilderStake",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"},
                {"name": "fee", "type": "uint256", "internalType": "uint256"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "initialize",
            "inputs": [
                {
                    "name": "clearing_",
                    "type": "address",
                    "internalType": "contract IProductRegistryFacet",
                }
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "maxPrice",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "int256", "internalType": "int256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "minPrice",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "int256", "internalType": "int256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "mmr",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "nextInterval",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "oracleSpecification",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct OracleSpecification",
                    "components": [
                        {
                            "name": "oracleAddress",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "fsvDecimals",
                            "type": "uint8",
                            "internalType": "uint8",
                        },
                        {
                            "name": "fspAlpha",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {"name": "fspBeta", "type": "int256", "internalType": "int256"},
                        {
                            "name": "fsvCalldata",
                            "type": "bytes",
                            "internalType": "bytes",
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "owner",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "pointValue",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "predictionProductV1",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct PredictionProductV1",
                    "components": [
                        {
                            "name": "base",
                            "type": "tuple",
                            "internalType": "struct BaseProduct",
                            "components": [
                                {
                                    "name": "metadata",
                                    "type": "tuple",
                                    "internalType": "struct ProductMetadata",
                                    "components": [
                                        {
                                            "name": "builder",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "symbol",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                        {
                                            "name": "description",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                    ],
                                },
                                {
                                    "name": "oracleSpec",
                                    "type": "tuple",
                                    "internalType": "struct OracleSpecification",
                                    "components": [
                                        {
                                            "name": "oracleAddress",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "fsvDecimals",
                                            "type": "uint8",
                                            "internalType": "uint8",
                                        },
                                        {
                                            "name": "fspAlpha",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fspBeta",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fsvCalldata",
                                            "type": "bytes",
                                            "internalType": "bytes",
                                        },
                                    ],
                                },
                                {
                                    "name": "collateralAsset",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "startTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pointValue",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "priceDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "extendedMetadata",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "expirySpec",
                            "type": "tuple",
                            "internalType": "struct ExpirySpecification",
                            "components": [
                                {
                                    "name": "earliestFSPSubmissionTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "tradeoutInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "maxPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "minPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "priceDecimals",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "productTreasury",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "products",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {"name": "", "type": "uint8", "internalType": "enum ProductType"},
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct BaseProduct",
                    "components": [
                        {
                            "name": "metadata",
                            "type": "tuple",
                            "internalType": "struct ProductMetadata",
                            "components": [
                                {
                                    "name": "builder",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "symbol",
                                    "type": "string",
                                    "internalType": "string",
                                },
                                {
                                    "name": "description",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "oracleSpec",
                            "type": "tuple",
                            "internalType": "struct OracleSpecification",
                            "components": [
                                {
                                    "name": "oracleAddress",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "fsvDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "fspAlpha",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "fspBeta",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "fsvCalldata",
                                    "type": "bytes",
                                    "internalType": "bytes",
                                },
                            ],
                        },
                        {
                            "name": "collateralAsset",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "startTime",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "pointValue",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "priceDecimals",
                            "type": "uint8",
                            "internalType": "uint8",
                        },
                        {
                            "name": "extendedMetadata",
                            "type": "string",
                            "internalType": "string",
                        },
                    ],
                },
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "proxiableUUID",
            "inputs": [],
            "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "registerFutureProduct",
            "inputs": [
                {
                    "name": "product",
                    "type": "tuple",
                    "internalType": "struct FuturesProductV1",
                    "components": [
                        {
                            "name": "base",
                            "type": "tuple",
                            "internalType": "struct BaseProduct",
                            "components": [
                                {
                                    "name": "metadata",
                                    "type": "tuple",
                                    "internalType": "struct ProductMetadata",
                                    "components": [
                                        {
                                            "name": "builder",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "symbol",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                        {
                                            "name": "description",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                    ],
                                },
                                {
                                    "name": "oracleSpec",
                                    "type": "tuple",
                                    "internalType": "struct OracleSpecification",
                                    "components": [
                                        {
                                            "name": "oracleAddress",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "fsvDecimals",
                                            "type": "uint8",
                                            "internalType": "uint8",
                                        },
                                        {
                                            "name": "fspAlpha",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fspBeta",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fsvCalldata",
                                            "type": "bytes",
                                            "internalType": "bytes",
                                        },
                                    ],
                                },
                                {
                                    "name": "collateralAsset",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "startTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pointValue",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "priceDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "extendedMetadata",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "expirySpec",
                            "type": "tuple",
                            "internalType": "struct ExpirySpecification",
                            "components": [
                                {
                                    "name": "earliestFSPSubmissionTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "tradeoutInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "marginSpec",
                            "type": "tuple",
                            "internalType": "struct MarginSpecification",
                            "components": [
                                {
                                    "name": "imr",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "mmr",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                    ],
                },
                {
                    "name": "initialBuilderStake",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "registerPredictionProduct",
            "inputs": [
                {
                    "name": "product",
                    "type": "tuple",
                    "internalType": "struct PredictionProductV1",
                    "components": [
                        {
                            "name": "base",
                            "type": "tuple",
                            "internalType": "struct BaseProduct",
                            "components": [
                                {
                                    "name": "metadata",
                                    "type": "tuple",
                                    "internalType": "struct ProductMetadata",
                                    "components": [
                                        {
                                            "name": "builder",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "symbol",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                        {
                                            "name": "description",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                    ],
                                },
                                {
                                    "name": "oracleSpec",
                                    "type": "tuple",
                                    "internalType": "struct OracleSpecification",
                                    "components": [
                                        {
                                            "name": "oracleAddress",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "fsvDecimals",
                                            "type": "uint8",
                                            "internalType": "uint8",
                                        },
                                        {
                                            "name": "fspAlpha",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fspBeta",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fsvCalldata",
                                            "type": "bytes",
                                            "internalType": "bytes",
                                        },
                                    ],
                                },
                                {
                                    "name": "collateralAsset",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "startTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pointValue",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "priceDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "extendedMetadata",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "expirySpec",
                            "type": "tuple",
                            "internalType": "struct ExpirySpecification",
                            "components": [
                                {
                                    "name": "earliestFSPSubmissionTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "tradeoutInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "maxPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "minPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                    ],
                },
                {
                    "name": "initialBuilderStake",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "registerPredictionProductFor",
            "inputs": [
                {
                    "name": "product",
                    "type": "tuple",
                    "internalType": "struct PredictionProductV1",
                    "components": [
                        {
                            "name": "base",
                            "type": "tuple",
                            "internalType": "struct BaseProduct",
                            "components": [
                                {
                                    "name": "metadata",
                                    "type": "tuple",
                                    "internalType": "struct ProductMetadata",
                                    "components": [
                                        {
                                            "name": "builder",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "symbol",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                        {
                                            "name": "description",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                    ],
                                },
                                {
                                    "name": "oracleSpec",
                                    "type": "tuple",
                                    "internalType": "struct OracleSpecification",
                                    "components": [
                                        {
                                            "name": "oracleAddress",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "fsvDecimals",
                                            "type": "uint8",
                                            "internalType": "uint8",
                                        },
                                        {
                                            "name": "fspAlpha",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fspBeta",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fsvCalldata",
                                            "type": "bytes",
                                            "internalType": "bytes",
                                        },
                                    ],
                                },
                                {
                                    "name": "collateralAsset",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "startTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pointValue",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "priceDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "extendedMetadata",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "expirySpec",
                            "type": "tuple",
                            "internalType": "struct ExpirySpecification",
                            "components": [
                                {
                                    "name": "earliestFSPSubmissionTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "tradeoutInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "maxPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "minPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                    ],
                },
                {
                    "name": "initialBuilderStake",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "removeAllowed",
            "inputs": [{"name": "addr", "type": "address", "internalType": "address"}],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "renounceOwnership",
            "inputs": [],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "state",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {"name": "", "type": "uint8", "internalType": "enum ProductState"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "tickSize",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "transferOwnership",
            "inputs": [
                {"name": "newOwner", "type": "address", "internalType": "address"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "upgradeToAndCall",
            "inputs": [
                {
                    "name": "newImplementation",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "data", "type": "bytes", "internalType": "bytes"},
            ],
            "outputs": [],
            "stateMutability": "payable",
        },
        {
            "type": "event",
            "name": "Initialized",
            "inputs": [
                {
                    "name": "version",
                    "type": "uint64",
                    "indexed": False,
                    "internalType": "uint64",
                }
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "OwnershipTransferred",
            "inputs": [
                {
                    "name": "previousOwner",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "newOwner",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "ProductRegistered",
            "inputs": [
                {
                    "name": "builder",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "productId",
                    "type": "bytes32",
                    "indexed": False,
                    "internalType": "bytes32",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "Upgraded",
            "inputs": [
                {
                    "name": "implementation",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                }
            ],
            "anonymous": False,
        },
        {
            "type": "error",
            "name": "AddressEmptyCode",
            "inputs": [
                {"name": "target", "type": "address", "internalType": "address"}
            ],
        },
        {
            "type": "error",
            "name": "ERC1967InvalidImplementation",
            "inputs": [
                {"name": "implementation", "type": "address", "internalType": "address"}
            ],
        },
        {"type": "error", "name": "ERC1967NonPayable", "inputs": []},
        {"type": "error", "name": "FailedCall", "inputs": []},
        {"type": "error", "name": "InvalidInitialization", "inputs": []},
        {
            "type": "error",
            "name": "InvalidParameter",
            "inputs": [
                {"name": "paramName", "type": "string", "internalType": "string"}
            ],
        },
        {"type": "error", "name": "NotInitializing", "inputs": []},
        {
            "type": "error",
            "name": "OwnableInvalidOwner",
            "inputs": [{"name": "owner", "type": "address", "internalType": "address"}],
        },
        {
            "type": "error",
            "name": "OwnableUnauthorizedAccount",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"}
            ],
        },
        {"type": "error", "name": "UUPSUnauthorizedCallContext", "inputs": []},
        {
            "type": "error",
            "name": "UUPSUnsupportedProxiableUUID",
            "inputs": [{"name": "slot", "type": "bytes32", "internalType": "bytes32"}],
        },
        {
            "type": "error",
            "name": "Unauthorized",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"}
            ],
        },
    ],
)



"""FinalSettlementFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import hexbytes
import web3
from web3 import types
from web3.contract import contract


class FinalSettlementFacet:
    """FinalSettlementFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed FinalSettlementFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def FSPFinalized(self) -> contract.ContractEvent:
        """Binding for `event FSPFinalized` on the FinalSettlementFacet contract."""
        return self._contract.events.FSPFinalized

    @property
    def FeeCollected(self) -> contract.ContractEvent:
        """Binding for `event FeeCollected` on the FinalSettlementFacet contract."""
        return self._contract.events.FeeCollected

    @property
    def FeeDispersed(self) -> contract.ContractEvent:
        """Binding for `event FeeDispersed` on the FinalSettlementFacet contract."""
        return self._contract.events.FeeDispersed

    @property
    def FinalSettlementCloseout(self) -> contract.ContractEvent:
        """Binding for `event FinalSettlementCloseout` on the FinalSettlementFacet contract."""
        return self._contract.events.FinalSettlementCloseout

    @property
    def PositionUpdated(self) -> contract.ContractEvent:
        """Binding for `event PositionUpdated` on the FinalSettlementFacet contract."""
        return self._contract.events.PositionUpdated

    def final_settlement_id(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `FINAL_SETTLEMENT_ID` on the FinalSettlementFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.FINAL_SETTLEMENT_ID().call(
            block_identifier=block_identifier
        )
        return int(return_value)

    def finalize_fsp(
        self,
        product_id: hexbytes.HexBytes,
    ) -> contract.ContractFunction:
        """Binding for `finalizeFsp` on the FinalSettlementFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.finalizeFsp(
            product_id,
        )

    def get_fsp(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.Tuple[int, bool]:
        """Binding for `getFsp` on the FinalSettlementFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        bool
        """
        return_value = self._contract.functions.getFsp(
            product_id,
        ).call(block_identifier=block_identifier)
        return (
            int(return_value[0]),
            bool(return_value[1]),
        )

    def get_fsp_finalization_time(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `getFspFinalizationTime` on the FinalSettlementFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.getFspFinalizationTime(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def initiate_final_settlement(
        self,
        product_id: hexbytes.HexBytes,
        accounts: typing.List[eth_typing.ChecksumAddress],
    ) -> contract.ContractFunction:
        """Binding for `initiateFinalSettlement` on the FinalSettlementFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        accounts : typing.List[eth_typing.ChecksumAddress]

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initiateFinalSettlement(
            product_id,
            accounts,
        )

    def open_interest(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `openInterest` on the FinalSettlementFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.openInterest(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "type": "function",
            "name": "FINAL_SETTLEMENT_ID",
            "inputs": [],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "finalizeFsp",
            "inputs": [
                {"name": "productID", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "int256", "internalType": "int256"}],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "getFsp",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [
                {"name": "fsp", "type": "int256", "internalType": "int256"},
                {"name": "finalized", "type": "bool", "internalType": "bool"},
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "getFspFinalizationTime",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "initiateFinalSettlement",
            "inputs": [
                {"name": "productID", "type": "bytes32", "internalType": "bytes32"},
                {"name": "accounts", "type": "address[]", "internalType": "address[]"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "openInterest",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "event",
            "name": "FSPFinalized",
            "inputs": [
                {
                    "name": "productID",
                    "type": "bytes32",
                    "indexed": True,
                    "internalType": "bytes32",
                },
                {
                    "name": "fsp",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "FeeCollected",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "capitalAmount",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "id",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "FeeDispersed",
            "inputs": [
                {
                    "name": "recipient",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "capitalAmount",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "id",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "FinalSettlementCloseout",
            "inputs": [
                {
                    "name": "productID",
                    "type": "bytes32",
                    "indexed": True,
                    "internalType": "bytes32",
                },
                {
                    "name": "accountLength",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
                {
                    "name": "closedBy",
                    "type": "address",
                    "indexed": False,
                    "internalType": "address",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "PositionUpdated",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "positionId",
                    "type": "bytes32",
                    "indexed": True,
                    "internalType": "bytes32",
                },
                {
                    "name": "costBasis",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "price",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "quantity",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "id",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "error",
            "name": "FSPAlreadyFinalized",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
        },
        {
            "type": "error",
            "name": "FSPNotFound",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
        },
        {
            "type": "error",
            "name": "FSPTimeNotReached",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"},
                {"name": "currentTime", "type": "uint256", "internalType": "uint256"},
                {
                    "name": "earliestFSPSubmissionTime",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
        },
        {
            "type": "error",
            "name": "InvalidFieldAccess",
            "inputs": [
                {
                    "name": "productType",
                    "type": "uint8",
                    "internalType": "enum ProductType",
                },
                {"name": "field", "type": "string", "internalType": "string"},
            ],
        },
        {
            "type": "error",
            "name": "InvalidOracleAddress",
            "inputs": [
                {"name": "oracleAddress", "type": "address", "internalType": "address"}
            ],
        },
        {
            "type": "error",
            "name": "InvalidParameter",
            "inputs": [
                {"name": "paramName", "type": "string", "internalType": "string"}
            ],
        },
        {
            "type": "error",
            "name": "MAECheckFailed",
            "inputs": [
                {"name": "marginAccount", "type": "address", "internalType": "address"}
            ],
        },
        {
            "type": "error",
            "name": "MismatchedFSPAccountQuantities",
            "inputs": [
                {"name": "checksum", "type": "int256", "internalType": "int256"},
                {
                    "name": "expectedChecksum",
                    "type": "int256",
                    "internalType": "int256",
                },
            ],
        },
        {
            "type": "error",
            "name": "NotFound",
            "inputs": [
                {"name": "parameter", "type": "string", "internalType": "string"}
            ],
        },
        {
            "type": "error",
            "name": "ProductNotInFinalSettlement",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
        },
        {
            "type": "error",
            "name": "SafeCastOverflowedUintToInt",
            "inputs": [{"name": "value", "type": "uint256", "internalType": "uint256"}],
        },
    ],
)



from decimal import Decimal
from functools import cache

from eth_typing.evm import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from web3.exceptions import ContractCustomError

from .. import validators
from ..bindings import (
    MarginAccount as MarginContract,
    MarginAccountRegistry,
    ProductRegistry,
)
from ..bindings.erc20 import ERC20
from ..bindings.facade import CLEARING_DIAMOND_ABI
from ..bindings.margin_account import ABI as MARGIN_CONTRACT_ABI
from ..bindings.margin_account_registry import ABI as MARGIN_ACCOUNT_REGISTRY_ABI
from ..decorators import convert_web3_error
from ..exceptions import NotFoundError
from ..schemas import Position, Transaction
from .base import ClearingSystemAPI


class MarginAccount(ClearingSystemAPI):
    """API for managing margin accounts."""

    ### Transactions ###

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def authorize(self, collateral_asset: str, intent_account_id: str) -> Transaction:
        """Authorizes a blockchain account to submit intents to the clearing system
        using the margin account associated with the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.
        intent_account_id : str
            The address of the intent account.

        Returns
        -------
        afp.schemas.Transaction
            Transaction parameters.
        """
        collateral_asset = validators.validate_address(collateral_asset)
        intent_account_id = validators.validate_address(intent_account_id)
        return self._transact(
            self._margin_contract(collateral_asset).authorize(intent_account_id)
        )

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def deposit(
        self, collateral_asset: str, amount: Decimal
    ) -> tuple[Transaction, Transaction]:
        """Deposits the specified amount of collateral tokens into the margin account
        associated with the collateral asset.

        First approves the token transfer with the collateral token, then executes the
        transfer.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.
        amount : Decimal
            The amount of collateral tokens to deposit.

        Returns
        -------
        afp.schemas.Transaction
            Parameters of the approval transaction.
        afp.schemas.Transaction
            Parameters of the deposit transaction.
        """
        collateral_asset = validators.validate_address(collateral_asset)
        token_amount = int(amount * 10 ** self._decimals(collateral_asset))
        token_contract = ERC20(self._w3, collateral_asset)

        tx1 = self._transact(
            token_contract.approve(
                self._margin_contract(collateral_asset)._contract.address,  # type: ignore
                token_amount,
            )
        )
        tx2 = self._transact(
            self._margin_contract(collateral_asset).deposit(token_amount)
        )
        return (tx1, tx2)

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def withdraw(self, collateral_asset: str, amount: Decimal) -> Transaction:
        """Withdraws the specified amount of collateral tokens from the margin account
        associated with the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.
        amount : Decimal
            The amount of collateral tokens to withdraw.

        Returns
        -------
        afp.schemas.Transaction
            Transaction parameters.
        """
        collateral_asset = validators.validate_address(collateral_asset)
        token_amount = int(amount * 10 ** self._decimals(collateral_asset))
        return self._transact(
            self._margin_contract(collateral_asset).withdraw(token_amount)
        )

    ### Views ###

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def capital(self, collateral_asset: str) -> Decimal:
        """Returns the amount of collateral tokens in the margin account associated
        with the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).capital(
            self._authenticator.address
        )
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def position(self, collateral_asset: str, position_id: str) -> Position:
        """Returns the parameters of a position in the margin account associated with
        the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.
        position_id: str
            The ID of the position.

        Returns
        -------
        afp.schemas.Position
        """
        validators.validate_hexstr32(position_id)
        data = self._margin_contract(collateral_asset).position_data(
            self._authenticator.address, HexBytes(position_id)
        )
        decimals = self._decimals(collateral_asset)
        return Position(
            id=Web3.to_hex(data.position_id),
            quantity=data.quantity,
            cost_basis=Decimal(data.cost_basis) / 10**decimals,
            maintenance_margin=Decimal(data.maintenance_margin) / 10**decimals,
            pnl=Decimal(data.pnl) / 10**decimals,
        )

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def positions(self, collateral_asset: str) -> list[Position]:
        """Returns all positions in the margin account associated with the collateral
        asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        list of afp.schemas.Position
        """
        collateral_asset = validators.validate_address(collateral_asset)
        position_ids = self._margin_contract(collateral_asset).positions(
            self._authenticator.address
        )
        return [self.position(collateral_asset, Web3.to_hex(id)) for id in position_ids]

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def equity(self, collateral_asset: str) -> Decimal:
        """Returns the margin account equity in the margin account associated with the
        collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).mae(
            self._authenticator.address
        )
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def maintenance_margin_available(self, collateral_asset: str) -> Decimal:
        """Returns the maintenance margin available in the margin account associated
        with the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).mma(
            self._authenticator.address
        )
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def maintenance_margin_used(self, collateral_asset: str) -> Decimal:
        """Returns the maintenance margin used in the margin account associated with
        the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).mmu(
            self._authenticator.address
        )
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def profit_and_loss(self, collateral_asset: str) -> Decimal:
        """Returns the profit and loss in the margin account associated with the
        collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).pnl(
            self._authenticator.address
        )
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(MARGIN_CONTRACT_ABI, CLEARING_DIAMOND_ABI)
    def withdrawable_amount(self, collateral_asset: str) -> Decimal:
        """Returns the amount of collateral tokens withdrawable from the margin account
        associated with the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).withdrawable(
            self._authenticator.address
        )
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    ### Internal getters ###

    @cache
    @convert_web3_error(MARGIN_ACCOUNT_REGISTRY_ABI)
    def _margin_contract(self, collateral_asset: ChecksumAddress) -> MarginContract:
        margin_account_registry_contract = MarginAccountRegistry(
            self._w3, self._config.margin_account_registry_address
        )
        try:
            margin_contract_address = (
                margin_account_registry_contract.get_margin_account(
                    Web3.to_checksum_address(collateral_asset)
                )
            )
        except ContractCustomError:
            raise NotFoundError("No margin account found for collateral asset")
        return MarginContract(self._w3, margin_contract_address)

    @cache
    def _tick_size(self, product_id: str) -> int:
        product_registry_contract = ProductRegistry(
            self._w3, self._config.product_registry_address
        )
        return product_registry_contract.tick_size(HexBytes(product_id))
